@@ -4,10 +4,11 @@
 // file that was distributed with this source code.
 // spell-checker:ignore (words) helloworld nodir objdump n'source
 
-use crate::common::util::{is_ci, TestScenario};
+use crate::common::util::{is_ci, run_ucmd_as_root, TestScenario};
 use filetime::FileTime;
-use std::os::unix::fs::PermissionsExt;
-#[cfg(not(any(windows, target_os = "freebsd")))]
+use std::fs;
+use std::os::unix::fs::{MetadataExt, PermissionsExt};
+#[cfg(not(windows))]
 use std::process::Command;
 #[cfg(any(target_os = "linux", target_os = "android"))]
 use std::thread::sleep;
@@ -609,13 +610,17 @@ fn test_install_copy_then_compare_file_with_extra_mode() {
 }
 
 const STRIP_TARGET_FILE: &str = "helloworld_installed";
-#[cfg(not(any(windows, target_os = "freebsd")))]
+#[cfg(all(not(windows), not(target_os = "freebsd")))]
 const SYMBOL_DUMP_PROGRAM: &str = "objdump";
-#[cfg(not(any(windows, target_os = "freebsd")))]
+#[cfg(target_os = "freebsd")]
+const SYMBOL_DUMP_PROGRAM: &str = "llvm-objdump";
+#[cfg(not(windows))]
 const STRIP_SOURCE_FILE_SYMBOL: &str = "main";
 
 fn strip_source_file() -> &'static str {
-    if cfg!(target_os = "macos") {
+    if cfg!(target_os = "freebsd") {
+        "helloworld_freebsd"
+    } else if cfg!(target_os = "macos") {
         "helloworld_macos"
     } else if cfg!(target_arch = "arm") || cfg!(target_arch = "aarch64") {
         "helloworld_android"
@@ -625,8 +630,7 @@ fn strip_source_file() -> &'static str {
 }
 
 #[test]
-// FixME: Freebsd fails on 'No such file or directory'
-#[cfg(not(any(windows, target_os = "freebsd")))]
+#[cfg(not(windows))]
 fn test_install_and_strip() {
     let scene = TestScenario::new(util_name!());
     let at = &scene.fixtures;
@@ -649,8 +653,7 @@ fn test_install_and_strip() {
 }
 
 #[test]
-// FixME: Freebsd fails on 'No such file or directory'
-#[cfg(not(any(windows, target_os = "freebsd")))]
+#[cfg(not(windows))]
 fn test_install_and_strip_with_program() {
     let scene = TestScenario::new(util_name!());
     let at = &scene.fixtures;
@@ -674,6 +677,102 @@ fn test_install_and_strip_with_program() {
     assert!(!stdout.contains(STRIP_SOURCE_FILE_SYMBOL));
 }
 
+#[cfg(all(unix, feature = "chmod"))]
+#[test]
+fn test_install_and_strip_with_program_hyphen() {
+    let scene = TestScenario::new(util_name!());
+
+    let at = &scene.fixtures;
+    let content = r#"#!/bin/sh
+    printf -- '%s\n' "$1" | grep '^[^-]'
+    "#;
+    at.write("no-hyphen", content);
+    scene.ccmd("chmod").arg("+x").arg("no-hyphen").succeeds();
+
+    at.touch("src");
+    scene
+        .ucmd()
+        .arg("-s")
+        .arg("--strip-program")
+        .arg("./no-hyphen")
+        .arg("--")
+        .arg("src")
+        .arg("-dest")
+        .succeeds()
+        .no_stderr()
+        .stdout_is("./-dest\n");
+
+    scene
+        .ucmd()
+        .arg("-s")
+        .arg("--strip-program")
+        .arg("./no-hyphen")
+        .arg("--")
+        .arg("src")
+        .arg("./-dest")
+        .succeeds()
+        .no_stderr()
+        .stdout_is("./-dest\n");
+}
+
+#[cfg(all(unix, feature = "chmod"))]
+#[test]
+fn test_install_on_invalid_link_at_destination() {
+    let scene = TestScenario::new(util_name!());
+
+    let at = &scene.fixtures;
+    at.mkdir("src");
+    at.mkdir("dest");
+    let src_dir = at.plus("src");
+    let dst_dir = at.plus("dest");
+
+    at.touch("test.sh");
+    at.symlink_file(
+        "/opt/FakeDestination",
+        &dst_dir.join("test.sh").to_string_lossy(),
+    );
+    scene.ccmd("chmod").arg("+x").arg("test.sh").succeeds();
+    at.symlink_file("test.sh", &src_dir.join("test.sh").to_string_lossy());
+
+    scene
+        .ucmd()
+        .current_dir(&src_dir)
+        .arg(src_dir.join("test.sh"))
+        .arg(dst_dir.join("test.sh"))
+        .succeeds()
+        .no_stderr()
+        .no_stdout();
+}
+
+#[cfg(all(unix, feature = "chmod"))]
+#[test]
+fn test_install_on_invalid_link_at_destination_and_dev_null_at_source() {
+    let scene = TestScenario::new(util_name!());
+
+    let at = &scene.fixtures;
+    at.mkdir("src");
+    at.mkdir("dest");
+    let src_dir = at.plus("src");
+    let dst_dir = at.plus("dest");
+
+    at.touch("test.sh");
+    at.symlink_file(
+        "/opt/FakeDestination",
+        &dst_dir.join("test.sh").to_string_lossy(),
+    );
+    scene.ccmd("chmod").arg("+x").arg("test.sh").succeeds();
+    at.symlink_file("test.sh", &src_dir.join("test.sh").to_string_lossy());
+
+    scene
+        .ucmd()
+        .current_dir(&src_dir)
+        .arg("/dev/null")
+        .arg(dst_dir.join("test.sh"))
+        .succeeds()
+        .no_stderr()
+        .no_stdout();
+}
+
 #[test]
 #[cfg(not(windows))]
 fn test_install_and_strip_with_invalid_program() {
@@ -1538,3 +1637,80 @@ fn test_install_compare_option() {
         .code_is(1)
         .stderr_contains("Options --compare and --strip are mutually exclusive");
 }
+
+#[test]
+// Matches part of tests/install/basic-1
+fn test_t_exist_dir() {
+    let scene = TestScenario::new(util_name!());
+    let at = &scene.fixtures;
+
+    let source1 = "file";
+    let target_dir = "sub4/";
+    let target_file = "sub4/file_exists";
+
+    at.touch(source1);
+    at.mkdir(target_dir);
+    at.touch(target_file);
+
+    scene
+        .ucmd()
+        .arg("-t")
+        .arg(target_file)
+        .arg("-Dv")
+        .arg(source1)
+        .fails()
+        .stderr_contains("failed to access 'sub4/file_exists': Not a directory");
+}
+
+#[test]
+fn test_target_file_ends_with_slash() {
+    let scene = TestScenario::new(util_name!());
+    let at = &scene.fixtures;
+
+    let source = "source_file";
+    let target_dir = "dir";
+    let target_file = "dir/target_file";
+    let target_file_slash = format!("{}/", target_file);
+
+    at.touch(source);
+    at.mkdir(target_dir);
+    at.touch(target_file);
+
+    scene
+        .ucmd()
+        .arg("-t")
+        .arg(target_file_slash)
+        .arg("-D")
+        .arg(source)
+        .fails()
+        .stderr_contains("failed to access 'dir/target_file/': Not a directory");
+}
+
+#[test]
+fn test_install_root_combined() {
+    let ts = TestScenario::new(util_name!());
+    let at = ts.fixtures.clone();
+    at.touch("a");
+    at.touch("c");
+
+    let run_and_check = |args: &[&str], target: &str, expected_uid: u32, expected_gid: u32| {
+        if let Ok(result) = run_ucmd_as_root(&ts, args) {
+            result.success();
+            assert!(at.file_exists(target));
+
+            let metadata = fs::metadata(at.plus(target)).unwrap();
+            assert_eq!(metadata.uid(), expected_uid);
+            assert_eq!(metadata.gid(), expected_gid);
+        } else {
+            print!("Test skipped; requires root user");
+        }
+    };
+
+    run_and_check(&["-Cv", "-o1", "-g1", "a", "b"], "b", 1, 1);
+    run_and_check(&["-Cv", "-o2", "-g1", "a", "b"], "b", 2, 1);
+    run_and_check(&["-Cv", "-o2", "-g2", "a", "b"], "b", 2, 2);
+
+    run_and_check(&["-Cv", "-o2", "c", "d"], "d", 2, 0);
+    run_and_check(&["-Cv", "c", "d"], "d", 0, 0);
+    run_and_check(&["-Cv", "c", "d"], "d", 0, 0);
+}
